"""

"""
from __future__ import annotations

import pgtrigger
from django.contrib.auth import get_user_model
from django.contrib.postgres.fields import HStoreField
from django.core import checks
from django.db import models
from django.db.models import ExpressionWrapper, F


class AuditTrail(models.Model):
    """
    An audit trail model mixin, no fields except for a timestamp
    perhaps should be added to tables that utilize this.

    If this data will be accessed frequently consider an IndexedAuditTrail instead.
    """

    before = HStoreField()
    after = HStoreField()

    class Meta:
        abstract = True


class IndexedAuditTrail(models.Model):
    """
    An indexed audit trail model mixin, no fields except for a timestamp
    perhaps should be added to tables that utilize this.
    """

    before = HStoreField(db_index=True)
    after = HStoreField(db_index=True)

    class Meta:
        abstract = True


TRIGGER_SQL = """
    INSERT INTO {meta.audit_table._meta.db_table}(before, after)
        SELECT hstore(old), hstore(new);
    RETURN new;
"""

CRUD_TRIGGERS = [
    pgtrigger.Trigger(
        name="track_history_create",
        level=pgtrigger.Row,
        when=pgtrigger.After,
        operation=pgtrigger.Insert,
        func=pgtrigger.Func(TRIGGER_SQL),
    ),
    pgtrigger.Trigger(
        name="track_history_update",
        level=pgtrigger.Row,
        when=pgtrigger.After,
        operation=pgtrigger.Update,
        func=pgtrigger.Func(TRIGGER_SQL),
    ),
    pgtrigger.Trigger(
        name="track_history_delete",
        level=pgtrigger.Row,
        when=pgtrigger.After,
        operation=pgtrigger.Delete,
        func=pgtrigger.Func(TRIGGER_SQL),
    ),
]


class AuditTrigger(models.Model):
    """ """

    class Meta:
        abstract = True
        triggers = CRUD_TRIGGERS
        audit_table = None

    @classmethod
    def check(cls, **kwargs):
        """

        :param kwargs:
        :return:
        """
        errors = super().check(**kwargs)
        errors.extend(cls._check_audit_table(**kwargs))
        errors.extend(cls._check_audit_triggers(**kwargs))
        return errors

    @classmethod
    def _check_audit_table(cls, **kwargs) -> list[checks.Error]:
        """

        :param kwargs:
        :return:
        """
        audit_table = cls.get_audit_table()
        if not audit_table:
            return [
                checks.Error(
                    "no audit table.",
                    hint=f"No audit table defined for {cls}, define an audit_table in the Meta class.",
                    obj=cls,
                    id="django_audimatic.E001",
                )
            ]
        return []

    @classmethod
    def _check_audit_triggers(cls, **kwargs) -> list[checks.Error]:
        """

        :param kwargs:
        :return:
        """
        audit_triggers = cls._get_triggers()
        if not audit_triggers:
            return [
                checks.Error(
                    "no triggers.",
                    hint=f"No triggers defined for {cls}, "
                    "does your Meta class inherit from `AuditTrigger.Meta`?",
                    obj=cls,
                    id="django_audimatic.E002",
                )
            ]
        return []

    @classmethod
    def get_audit_table(cls) -> None | models.Model:
        """

        :return:
        """
        return getattr(cls._meta, "audit_table", None)

    @classmethod
    def _get_triggers(cls) -> list:
        """

        :return:
        """
        return getattr(cls._meta, "triggers", None)

    def get_audit_trail(self) -> models.QuerySet:
        """

        :return:
        """
        audit_table = self.get_audit_table()
        pk = self.id
        return (
            (
                audit_table.objects.filter(before__id__contains=pk)
                | audit_table.objects.filter(after__id__contains=pk)
            )
            .annotate(
                diff=ExpressionWrapper(
                    F("after") - F("before"), output_field=HStoreField()
                )
            )
            .all()
        )


class AuditActions(models.Model):
    """Tracks actions performed on a table.

    When a record is restored from an audit trail, etc.
    """

    action = models.CharField(max_length=128)
    audit_table = models.CharField(max_length=128)
    audit_row_id = models.BigIntegerField()
    timestamp = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(
        get_user_model(), on_delete=models.DO_NOTHING, null=True, blank=True
    )



# Generated by Django 4.2.5 on 2023-09-21 03:29

import django.db.models.deletion
from django.conf import settings
from django.contrib.postgres.operations import HStoreExtension
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        HStoreExtension(),
        migrations.CreateModel(
            name="AuditActions",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("action", models.CharField(max_length=128)),
                ("audit_table", models.CharField(max_length=128)),
                ("audit_row_id", models.BigIntegerField()),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



from django.contrib import admin
from django.contrib.postgres.fields import HStoreField
from django.db.models import ExpressionWrapper, F
from nonrelated_inlines.admin import NonrelatedStackedInline

from django_audimatic.models import AuditTrigger


class DiffInline(NonrelatedStackedInline):
    fields = ["audit_report"]
    readonly_fields = ["audit_report"]
    max_num = 0
    can_delete = False

    @admin.display(description="Diff")
    def audit_report(self, instance):
        return instance.diff

    def get_form_queryset(self, obj: AuditTrigger):
        return (
            (
                self.model.objects.filter(before__id__contains=obj.id)
                | self.model.objects.filter(after__id__contains=obj.id)
            )
            .annotate(
                diff=ExpressionWrapper(
                    F("after") - F("before"), output_field=HStoreField()
                )
            )
            .all()
        )

    def save_new_instance(self, parent, instance):
        raise NotImplementedError("Saving is done via postgresql triggers.")



"""Then steps."""
from behave import then


@then('the "{model_name}" model should have an audit trail')
def step_impl(context, model_name):
    """Check that the model has an audit trail."""
    model = context.models[model_name]
    context.test.assertTrue(hasattr(model, "get_audit_trail"))



from collections import OrderedDict

from django.apps import AppConfig
from django.db.migrations import state
from django.db.models import options

if "triggers" not in state.DEFAULT_NAMES:  # pragma: no branch
    state.DEFAULT_NAMES = tuple(state.DEFAULT_NAMES) + ("triggers",)
if "triggers" not in options.DEFAULT_NAMES:  # pragma: no branch
    options.DEFAULT_NAMES = tuple(options.DEFAULT_NAMES) + ("triggers",)

if "audit_table" not in options.DEFAULT_NAMES:  # pragma: no branch
    options.DEFAULT_NAMES = tuple(options.DEFAULT_NAMES) + (
        "audit_table",
        "audit_options",
    )
if "audit_table" not in state.DEFAULT_NAMES:  # pragma: no branch
    state.DEFAULT_NAMES = tuple(state.DEFAULT_NAMES) + (
        "audit_table",
        "audit_options",
    )


def patch_migrations():
    """
    Patch the autodetector and model state detection if migrations are turned on
    """
    if "triggers" not in state.DEFAULT_NAMES:  # pragma: no branch
        state.DEFAULT_NAMES = tuple(state.DEFAULT_NAMES) + ("triggers",)

    if "audit_table" not in state.DEFAULT_NAMES:  # pragma: no branch
        state.DEFAULT_NAMES = tuple(state.DEFAULT_NAMES) + (
            "audit_table",
            "audit_options",
        )

    if "triggers" not in options.DEFAULT_NAMES:  # pragma: no branch
        options.DEFAULT_NAMES = tuple(options.DEFAULT_NAMES) + ("triggers",)
    if "audit_table" not in options.DEFAULT_NAMES:  # pragma: no branch
        options.DEFAULT_NAMES = tuple(options.DEFAULT_NAMES) + (
            "audit_table",
            "audit_options",
        )


class DjangoAudimaticConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_audimatic"

    def ready(self):
        """
        Do all necessary patching
        """

        from django.apps import apps
        from django.conf import settings

        patch_migrations()
        dirty = False

        if "pgtrigger" not in settings.INSTALLED_APPS:
            settings.INSTALLED_APPS.append("pgtrigger")
            dirty = True

        if "nonrelated_inlines" not in settings.INSTALLED_APPS:
            settings.INSTALLED_APPS.append("nonrelated_inlines")
            dirty = True

        if dirty:
            apps.app_configs = OrderedDict()
            apps.apps_ready = apps.models_ready = apps.loading = apps.ready = False
            apps.clear_cache()
            apps.populate(settings.INSTALLED_APPS)
